_lazy_model_map = {
    "Dire": "ForensicHub.tasks.aigc.models.dire.dire.dire",
    "HiFi_Net": "ForensicHub.tasks.aigc.models.hifi_net.HiFi_Net",
    "DualNet": "ForensicHub.tasks.aigc.models.dualnet.DualNet",
    "Synthbuster": "ForensicHub.tasks.aigc.models.synthbuster.Synthbuster",
    "UnivFD": "ForensicHub.tasks.aigc.models.univfd.UnivFD",
    "FatFormer": "ForensicHub.tasks.aigc.models.fatformer.FatFormer",
    "CO_SPY": "ForensicHub.tasks.aigc.models.co_spy.CO_SPY",
}

_lazy_postfunc_map = {

}
